import { Promise } from 'bluebird';
import { URL } from 'url';

import Knex = require('knex');
import Timeout = NodeJS.Timer;

/* eslint-disable no-param-reassign */

type InfoLogger = (input: { message: string, data?: Object }) => void;
type ErrorLogger = (input:
                      { message: string, error?: Error, data?: Object } |
                      { message?: string, error: Error, data?: Object }) => void;

export interface PDConstructorArgs {
  connections: string[];
  options: PDOptions;
  knexOptions?: Knex.Config;
  logger?: {
    info?: InfoLogger;
    error?: ErrorLogger;
  };
}

export interface PDConnecton {
  host: string;
  knex: Knex;
  primary: boolean;
  healthy: boolean;
  latency: number;
}

interface PDCheckupResult {
  pingOk: boolean;
  primary: boolean;
}

export interface PDOptions {
  healthcheckInterval: number;
  healthcheckTimeout: number;
}

export class PDError extends Error {
  public code: string;
}

const infoLogger: InfoLogger = function infoLogger(input) {
  console.log(input.message + '\n', input.data || {});
};

const errorLogger: ErrorLogger = function errorLogger(input) {
  const args = [];
  if (input.message) {
    args.push(input.message + '\n');
  }
  if (input.error) {
    args.push(input.error);
  }
  if (input.data) {
    args.push(input.data);
  }
  console.error(...args);
};

export class PGDispatcher {
  private connections: PDConnecton[];

  private options: PDOptions;

  private logger: { info: InfoLogger, error: ErrorLogger };

  private hcTimer: Timeout | null;

  constructor({
    connections = [],
    options,
    knexOptions = {},
    logger = {},
  }: PDConstructorArgs) {
    if (!connections.length) {
      throw new Error('Empty connections list is not allowed');
    }

    this.connections = connections.map((connectionString) => {
      const url = new URL(connectionString);
      return {
        host: url.hostname,
        knex: Knex({ // eslint-disable-line new-cap
          client: 'pg',
          connection: connectionString,
          ...knexOptions,
        }),
        primary: false,
        healthy: false,
        latency: Infinity,
      };
    });
    this.options = options;

    this.logger = {
      info: logger.info || infoLogger,
      error: logger.error || errorLogger,
    };

    this.initHealthcheck().catch((error) => {
      this.logger.error({ error });
    });
  }

  private async initHealthcheck() {
    await this.knexReady();

    this.hcTimer = setInterval(() => {
      const checkups = this.connections.map(connection => this.checkDatabase(connection)
        .catch((error) => {
          this.logger.error({ message: 'Unhandled error during database checkup', error });
          connection.healthy = false;
          connection.primary = false;
        }));
      Promise.all(checkups).then(() => {
        this.logger.info({
          message: 'Database current status',
          data: {
            connections: this.connections.map(c => ({
              host: c.host,
              primary: c.primary,
              healthy: c.healthy,
              latency: c.latency,
            })),
          },
        });
      });
    }, this.options.healthcheckInterval);
  }

  private async performCheckupQuery(knex: Knex): Promise<PDCheckupResult> {
    const result = await knex
      .raw('SELECT pg_is_in_recovery();')
      .timeout(this.options.healthcheckTimeout);

    const isReplica = result
      && result.rows
      && result.rows[0]
      && result.rows[0].pg_is_in_recovery;

    const pingOk = typeof isReplica === 'boolean';
    const isPrimary = pingOk && isReplica === false;

    return { pingOk, primary: isPrimary };
  }

  private async checkDatabase(connection: PDConnecton) {
    try {
      const startTime = new Date().getTime();
      const checkupResult = await Promise.any([
        Promise.resolve(this.performCheckupQuery(connection.knex)).catch(() => {
        }),
        Promise.delay(this.options.healthcheckTimeout + 10),
      ]);
      connection.latency = new Date().getTime() - startTime;

      if (checkupResult) {
        connection.healthy = checkupResult.pingOk;
        connection.primary = checkupResult.primary;
      } else {
        this.logger.error({ message: 'Database checkup timeouted', data: { host: connection.host } });
        connection.healthy = false;
        connection.primary = false;
      }
    } catch (error) {
      connection.healthy = false;
      connection.primary = false;
      this.logger.error({
        message: 'Database checkup failed',
        error,
        data: { host: connection.host },
      });
    }
  }

  private async knexReady() {
    await Promise.all(this.connections.map(c => c.knex));
  }

  async ready() {
    await this.knexReady();
    await Promise.all(this.connections.map(c => this.checkDatabase(c)));
  }

  terminate() {
    if (this.hcTimer) {
      clearInterval(this.hcTimer);
    }

    return Promise.all(this.connections.map(c => c.knex.destroy()
      .then(() => {
        this.logger.info({ message: 'Connection successfully terminated', data: { host: c.host } });
      })
      .catch(error => this.logger.error({
        message: 'Error during connection terminating',
        error,
        data: { host: c.host },
      }))));
  }

  private get healthyConnections() {
    return this.connections.filter(c => c.healthy);
  }

  private checkConnectionsAvailability() {
    if (!this.healthyConnections.length) {
      const error = new PDError('No connections available');
      error.code = 'ERR_DB_NOT_AVAILABLE';
      throw error;
    }
  }

  get primary() {
    this.checkConnectionsAvailability();

    const primaryConnections = this.connections.filter(c => c.primary);
    if (primaryConnections.length > 1) {
      this.logger.error({
        message: 'Multiple primary connections detected, something is wrong',
        data: {
          primaryConnections: primaryConnections.map(c => c.host),
        },
      });
    }

    const primaryConnection = primaryConnections[0];
    if (!primaryConnection) {
      const error = new PDError('No primary connection available');
      error.code = 'ERR_DB_READ_ONLY';
      throw error;
    }
    return primaryConnection.knex;
  }

  get replica() {
    this.checkConnectionsAvailability();

    const replicaConnections = this.healthyConnections
      .filter(c => !c.primary);

    if (replicaConnections.length) {
      const fastestReplica = replicaConnections.sort((a, b) => a.latency - b.latency)[0];
      return fastestReplica.knex;
    }
    if (this.connections.length > 1) {
      this.logger.error({ message: 'No alive replica available, using master for read' });
    }
    return this.connections[0].knex;
  }
}
