import { log, reportError } from '../core/utils';
import { AppTask } from '../core/worker';

import AWS = require('aws-sdk');
import Bluebird = require('bluebird');

const sqs = new AWS.SQS({ apiVersion: '2012-11-05' });

export const Queues = [
  'common',
  'import',
  'critical',
];

const MAXIMUM_BATCH_SIZE = 10;
const SQS_BATCH_ADD_CONCURRENCY = 4;
const MessageGroupId = 'common';

type QueueUrlsMap = { [key: string]: string };

export const QueuesUrls: QueueUrlsMap = Queues.reduce((result: QueueUrlsMap, name) => {
  const envVariableName = `AWS_SQS_QUEUE_${name.toUpperCase()}`;
  const url = process.env[envVariableName];
  if (!url) {
    throw new Error(`Url is not set for queue "${name}" (${envVariableName})`);
  }
  result[name] = url; // eslint-disable-line
  return result;
}, {});

function getRandomId() {
  return String(Math.round(Math.random() * 1000000000));
}

export function enqueue(task: string, args: Array<string | number>, queue: string = 'common'): Promise<any> {
  const queueUrl = QueuesUrls[queue];
  if (!queueUrl) {
    throw new Error(`Unknown queue "${queue}"`);
  }

  const params = {
    MessageBody: JSON.stringify({
      task,
      args,
    }),
    MessageGroupId,
    QueueUrl: queueUrl,
  };

  return new Promise((resolve, reject) => {
    sqs.sendMessage(params, (error, data) => {
      if (error) {
        reportError(error, { message: 'Error during task sending' });
        reject(error);
        return;
      }
      resolve(data);
    });
  });
}

function splitBatches(input: any[], batchSize: number) {
  const batches = [];

  for (let i = 0, j = input.length; i < j; i += batchSize) {
    batches.push(input.slice(i, i + batchSize));
  }

  return batches;
}

function enqueueBatch(tasks: AppTask[], queueUrl: string): Promise<any> {
  const params = {
    Entries: tasks.map(t => ({
      Id: getRandomId(),
      MessageBody: JSON.stringify({
        task: t.task,
        args: t.args,
      }),
      MessageGroupId,
    })),
    QueueUrl: queueUrl,
  };

  return new Promise((resolve, reject) => {
    sqs.sendMessageBatch(params, (error, data) => {
      if (error) {
        reportError(error, { message: 'Error during task sending' });
        reject(error);
        return;
      }

      log('Successfully added tasks batch to the queue');
      resolve(data);
    });
  });
}

export function batchedEnqueue(tasks: AppTask[], queue: string = 'common'): Bluebird<any> {
  const queueUrl = QueuesUrls[queue];
  if (!queueUrl) {
    throw new Error(`Unknown queue "${queue}"`);
  }

  const batches = splitBatches(tasks, MAXIMUM_BATCH_SIZE);

  return Bluebird.map(batches, tasksBatch => enqueueBatch(tasksBatch, queueUrl), { concurrency: SQS_BATCH_ADD_CONCURRENCY });
}
