"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const bluebird_1 = require("bluebird");
const url_1 = require("url");
const Knex = require("knex");
class PDError extends Error {
}
exports.PDError = PDError;
const infoLogger = function infoLogger(input) {
    console.log(input.message + '\n', input.data || {});
};
const errorLogger = function errorLogger(input) {
    const args = [];
    if (input.message) {
        args.push(input.message + '\n');
    }
    if (input.error) {
        args.push(input.error);
    }
    if (input.data) {
        args.push(input.data);
    }
    console.error(...args);
};
class PGDispatcher {
    constructor({ connections = [], options, knexOptions = {}, logger = {}, }) {
        if (!connections.length) {
            throw new Error('Empty connections list is not allowed');
        }
        this.connections = connections.map((connectionString) => {
            const url = new url_1.URL(connectionString);
            return {
                host: url.hostname,
                knex: Knex(Object.assign({ client: 'pg', connection: connectionString }, knexOptions)),
                primary: false,
                healthy: false,
                latency: Infinity,
            };
        });
        this.options = options;
        this.logger = {
            info: logger.info || infoLogger,
            error: logger.error || errorLogger,
        };
        this.initHealthcheck().catch((error) => {
            this.logger.error({ error });
        });
    }
    async initHealthcheck() {
        await this.knexReady();
        this.hcTimer = setInterval(() => {
            const checkups = this.connections.map(connection => this.checkDatabase(connection)
                .catch((error) => {
                this.logger.error({ message: 'Unhandled error during database checkup', error });
                connection.healthy = false;
                connection.primary = false;
            }));
            bluebird_1.Promise.all(checkups).then(() => {
                this.logger.info({
                    message: 'Database current status',
                    data: {
                        connections: this.connections.map(c => ({
                            host: c.host,
                            primary: c.primary,
                            healthy: c.healthy,
                            latency: c.latency,
                        })),
                    },
                });
            });
        }, this.options.healthcheckInterval);
    }
    async performCheckupQuery(knex) {
        const result = await knex
            .raw('SELECT pg_is_in_recovery();')
            .timeout(this.options.healthcheckTimeout);
        const isReplica = result
            && result.rows
            && result.rows[0]
            && result.rows[0].pg_is_in_recovery;
        const pingOk = typeof isReplica === 'boolean';
        const isPrimary = pingOk && isReplica === false;
        return { pingOk, primary: isPrimary };
    }
    async checkDatabase(connection) {
        try {
            const startTime = new Date().getTime();
            const checkupResult = await bluebird_1.Promise.any([
                bluebird_1.Promise.resolve(this.performCheckupQuery(connection.knex)).catch(() => {
                }),
                bluebird_1.Promise.delay(this.options.healthcheckTimeout + 10),
            ]);
            connection.latency = new Date().getTime() - startTime;
            if (checkupResult) {
                connection.healthy = checkupResult.pingOk;
                connection.primary = checkupResult.primary;
            }
            else {
                this.logger.error({ message: 'Database checkup timeouted', data: { host: connection.host } });
                connection.healthy = false;
                connection.primary = false;
            }
        }
        catch (error) {
            connection.healthy = false;
            connection.primary = false;
            this.logger.error({
                message: 'Database checkup failed',
                error,
                data: { host: connection.host },
            });
        }
    }
    async knexReady() {
        await bluebird_1.Promise.all(this.connections.map(c => c.knex));
    }
    async ready() {
        await this.knexReady();
        await bluebird_1.Promise.all(this.connections.map(c => this.checkDatabase(c)));
    }
    terminate() {
        if (this.hcTimer) {
            clearInterval(this.hcTimer);
        }
        return bluebird_1.Promise.all(this.connections.map(c => c.knex.destroy()
            .then(() => {
            this.logger.info({ message: 'Connection successfully terminated', data: { host: c.host } });
        })
            .catch(error => this.logger.error({
            message: 'Error during connection terminating',
            error,
            data: { host: c.host },
        }))));
    }
    get healthyConnections() {
        return this.connections.filter(c => c.healthy);
    }
    checkConnectionsAvailability() {
        if (!this.healthyConnections.length) {
            const error = new PDError('No connections available');
            error.code = 'ERR_DB_NOT_AVAILABLE';
            throw error;
        }
    }
    get primary() {
        this.checkConnectionsAvailability();
        const primaryConnections = this.connections.filter(c => c.primary);
        if (primaryConnections.length > 1) {
            this.logger.error({
                message: 'Multiple primary connections detected, something is wrong',
                data: {
                    primaryConnections: primaryConnections.map(c => c.host),
                },
            });
        }
        const primaryConnection = primaryConnections[0];
        if (!primaryConnection) {
            const error = new PDError('No primary connection available');
            error.code = 'ERR_DB_READ_ONLY';
            throw error;
        }
        return primaryConnection.knex;
    }
    get replica() {
        this.checkConnectionsAvailability();
        const replicaConnections = this.healthyConnections
            .filter(c => !c.primary);
        if (replicaConnections.length) {
            const fastestReplica = replicaConnections.sort((a, b) => a.latency - b.latency)[0];
            return fastestReplica.knex;
        }
        if (this.connections.length > 1) {
            this.logger.error({ message: 'No alive replica available, using master for read' });
        }
        return this.connections[0].knex;
    }
}
exports.PGDispatcher = PGDispatcher;
//# sourceMappingURL=pg-dispatcher.js.map