"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("../core/utils");
const AWS = require("aws-sdk");
const Bluebird = require("bluebird");
const sqs = new AWS.SQS({ apiVersion: '2012-11-05' });
exports.Queues = [
    'common',
    'import',
    'critical',
];
const MAXIMUM_BATCH_SIZE = 10;
const SQS_BATCH_ADD_CONCURRENCY = 4;
const MessageGroupId = 'common';
exports.QueuesUrls = exports.Queues.reduce((result, name) => {
    const envVariableName = `AWS_SQS_QUEUE_${name.toUpperCase()}`;
    const url = process.env[envVariableName];
    if (!url) {
        throw new Error(`Url is not set for queue "${name}" (${envVariableName})`);
    }
    result[name] = url; // eslint-disable-line
    return result;
}, {});
function getRandomId() {
    return String(Math.round(Math.random() * 1000000000));
}
function enqueue(task, args, queue = 'common') {
    const queueUrl = exports.QueuesUrls[queue];
    if (!queueUrl) {
        throw new Error(`Unknown queue "${queue}"`);
    }
    const params = {
        MessageBody: JSON.stringify({
            task,
            args,
        }),
        MessageGroupId,
        QueueUrl: queueUrl,
    };
    return new Promise((resolve, reject) => {
        sqs.sendMessage(params, (error, data) => {
            if (error) {
                utils_1.reportError(error, { message: 'Error during task sending' });
                reject(error);
                return;
            }
            resolve(data);
        });
    });
}
exports.enqueue = enqueue;
function splitBatches(input, batchSize) {
    const batches = [];
    for (let i = 0, j = input.length; i < j; i += batchSize) {
        batches.push(input.slice(i, i + batchSize));
    }
    return batches;
}
function enqueueBatch(tasks, queueUrl) {
    const params = {
        Entries: tasks.map(t => ({
            Id: getRandomId(),
            MessageBody: JSON.stringify({
                task: t.task,
                args: t.args,
            }),
            MessageGroupId,
        })),
        QueueUrl: queueUrl,
    };
    return new Promise((resolve, reject) => {
        sqs.sendMessageBatch(params, (error, data) => {
            if (error) {
                utils_1.reportError(error, { message: 'Error during task sending' });
                reject(error);
                return;
            }
            utils_1.log('Successfully added tasks batch to the queue');
            resolve(data);
        });
    });
}
function batchedEnqueue(tasks, queue = 'common') {
    const queueUrl = exports.QueuesUrls[queue];
    if (!queueUrl) {
        throw new Error(`Unknown queue "${queue}"`);
    }
    const batches = splitBatches(tasks, MAXIMUM_BATCH_SIZE);
    return Bluebird.map(batches, tasksBatch => enqueueBatch(tasksBatch, queueUrl), { concurrency: SQS_BATCH_ADD_CONCURRENCY });
}
exports.batchedEnqueue = batchedEnqueue;
//# sourceMappingURL=queues.js.map